"""SSD1309 demo (fonts)."""
from select import select
from socket import socket, AF_INET, SOCK_DGRAM
from sys import modules
from time import sleep
from machine import Pin, SPI, Timer
from network import WLAN, STA_IF
from ssd1309 import Display
from xglcd_font import XglcdFont

PORT = const(3520)  # Caller ID broadcasts on UDP port 3520
BUFFER_SIZE = const(128)  # Maximum broadcast packet < 128 bytes
DISPLAY_WIDTH = const(128)  # Width of LCD display
DISPLAY_DELAY = const(60000)  # Display caller ID for 1 minute

PHONE_LINE = const(0)
CALL_TYPE = const(1)  # I=Incoming, O=Outgoing
INDICATOR = const(2)  # S or B=Call Start, E=Call End
CALL_DURATION = const(3)
CHECKSUM = const(4)  # G=Good, B=Bad
RING_COUNT = const(5)
CALL_DATE = const(6)
CALL_TIME = const(7)
CALL_AMPM = const(8)
PHONE_NUMBER = const(9)
CALLER_NAME = const(10)

spi = SPI(1, baudrate=10000000, sck=Pin(14), mosi=Pin(13))
display = Display(spi, dc=Pin(17), cs=Pin(5), rst=Pin(16))
display.clear()

wlan = WLAN(STA_IF)  # Display IP address while loading
if wlan.active():
    display.draw_text8x8(0, 55, wlan.ifconfig()[0])
else:
    display.draw_text8x8(0, 55, "Invalid IP")

display.draw_text8x8(0, 0, "Please wait,")
display.draw_text8x8(0, 16, "Loading fonts...")
display.present()

# Load fonts (takes a few seconds)
perfect_big = XglcdFont('fonts/PerfectPixel_23x32.c', 23, 32)
perfect_small = XglcdFont('fonts/PerfectPixel_18x25.c', 18, 25)
display.clear()
display.sleep()

contacts = {}  # Contacts dict
timer = Timer(1)  # Instantiate hardware timer


def load_contacts():
    """Load contacts from CSV file."""
    with open('contacts.csv', 'r') as file:
        for contact in file:
            if not contact:
                return  # Break at end of text file
            contact = contact.rstrip('\n')
            contact = contact.rstrip('\r')
            contact = contact.split(',')  # Split number & name
            # Add to dict
            contacts[contact[0].strip()] = contact[1].strip()


def display_name_max(name, x, y):
    """Display name on LCD and maximize font size."

    Args:
        name (string):  name to display on LCD
        x, y (int): coordinates to draw text
    """
    if perfect_big.measure_text(name) < DISPLAY_WIDTH:
        display.draw_text(x, y, name, perfect_big)
    else:
        display.draw_text(x, y, name, perfect_small)


def handle_ftp():
    timer.deinit()
    display.wake()
    display.clear()
    display.draw_text8x8(0, 16, "FTP Running")
    display.draw_text8x8(0, 32, wlan.ifconfig()[0])
    display.present()
    if 'ftp' not in modules:
        import ftp
    else:
        modules['ftp'].ftpserver()
    display.clear()
    display.sleep()


def timer_callback(t):
    """Clear display on hardware timer time out."""
    display.clear()
    display.sleep()
    t.deinit()


def parse_callerid(msg):
    """Parse caller ID broadcast data."

    Args:
        msg (bytes):  caller ID message bytes.
    """

    if len(msg) < 20:
        print("invalid message")
        return  # invalid message

    msg = msg[21:].strip()  # Remove header bytes
    msg = msg.decode("utf-8")  # Decode bytes to string

    # Extract call data and remove any blank spaces
    call_data = list(filter(None, msg.split(" ")))

    if len(call_data) < PHONE_NUMBER:
        print("Non-callerid broadcast")
        return

    try:
        if call_data[CALL_TYPE] != "I":
            print("invalid call type: {0}".format(
                  call_data[CALL_TYPE]))
            return  # Only care about incoming calls

        # B is not documented, but indicates
        # good caller when when blocking enabled
        # Otherwise S=Start of call and E=End of call
        if call_data[INDICATOR] != "B":
            return

        print("Phone Line: {0}".format(call_data[PHONE_LINE]))
        print("Call Type: {0}".format(call_data[CALL_TYPE]))
        print("Indicator: {0}".format(call_data[INDICATOR]))
        print("Duration: {0}".format(call_data[CALL_DURATION]))
        print("Checksum: {0}".format(call_data[CHECKSUM]))
        print("Ring Count: {0}".format(call_data[RING_COUNT]))
        print("Call Date: {0}".format(call_data[CALL_DATE]))
        print("Call Time: {0} {1}".format(call_data[CALL_TIME],
                                          call_data[CALL_AMPM]))
        print("Phone Number: {0}".format(call_data[PHONE_NUMBER]))
        caller_name = "NONE"
        if len(call_data) > CALLER_NAME:  # Check if name provided
            caller_name = call_data[CALLER_NAME]
            # Construct caller name if multiple names
            for i in range(CALLER_NAME + 1, len(call_data)):
                caller_name += " " + call_data[i]
        print("Caller Name: {0}".format(caller_name))

        timer.deinit()
        display.wake()
        display.clear()
        if call_data[PHONE_NUMBER] in contacts.keys():
            # If contact exists return caller name from number
            # Split first & last names
            names = contacts[call_data[PHONE_NUMBER]].split(" ")
            display_name_max(names[0], 0, 0)  # Display first name
            if len(names) > 1:  # Last name exists
                display_name_max(names[1], 0, 32)  # Display last
        else:
            # Unknown number erroneously got past caller id block
            display.draw_text(0, 0, "Do Not", perfect_big)
            display.draw_text(0, 32, "Answer", perfect_big)
        display.present()
        timer.init(mode=Timer.ONE_SHOT, period=DISPLAY_DELAY,
                   callback=timer_callback)

    except Exception as e:
        timer.deinit()
        display.wake()
        display.clear()
        if hasattr(e, 'message'):
            print(e.message)
            display.draw_text8x8(0, 0, e.message)
        else:
            print(str(e))
            display.draw_text8x8(0, 0, str(e))
        display.present()


def main():
    """Monitor Caller ID."""
    load_contacts()  # Load contacts from flash

    s = socket(AF_INET, SOCK_DGRAM)
    s.bind(('', PORT))
    s.setblocking(0)  # Disable blocking

    print("Waiting for calls...")

    while True:
        # Monitor socket until communication
        result = select([s], [], [])
        msg = result[0][0].recv(BUFFER_SIZE)
        print(msg)
        if msg == b'^^Id-f' or msg == b'^^Id-F':  # FTP command
            handle_ftp()
        else:
            parse_callerid(msg)


main()
